/**
 * ==============================================
 *  DQL : Descartes Query Language
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Juergen Walter, Fabian Gorsler, Fabian Brosig and other contributors.
 *
 * Project Info:   http://descartes.tools/dql
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.core.engine.interpreter;

import static tools.descartes.dql.core.engine.util.QueryParserHelper.stringToModel;
import static tools.descartes.dql.core.engine.util.QueryParserHelper.toFlatConfigurationPropertiesClause;

import java.util.HashMap;
import java.util.List;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;

import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;
import tools.descartes.dql.core.engine.util.SLAReporter;
import tools.descartes.dql.lang.descartesQL.ConfigurationPropertiesClause;
import tools.descartes.dql.lang.descartesQL.DescartesQL;
import tools.descartes.dql.lang.descartesQL.DoFReference;
import tools.descartes.dql.lang.descartesQL.DoFVariationClause;
import tools.descartes.dql.lang.descartesQL.DoubleValueVariationClause;
import tools.descartes.dql.lang.descartesQL.ExplorationStrategyClause;
import tools.descartes.dql.lang.descartesQL.Goal;
import tools.descartes.dql.lang.descartesQL.IntervalVariationClause;
import tools.descartes.dql.lang.descartesQL.MetricClause;
import tools.descartes.dql.lang.descartesQL.MetricClauseType;
import tools.descartes.dql.lang.descartesQL.ModelFamily;
import tools.descartes.dql.lang.descartesQL.ModelLocation;
import tools.descartes.dql.lang.descartesQL.ModelReferenceClause;
import tools.descartes.dql.lang.descartesQL.PerformanceMetricsQuery;
import tools.descartes.dql.lang.descartesQL.Sla;
import tools.descartes.dql.lang.descartesQL.Slo;
import tools.descartes.dql.lang.descartesQL.UsingClause;
import tools.descartes.dql.lang.descartesQL.ValueVariationClause;
import tools.descartes.dql.lang.descartesQL.WeightedSlo;
import tools.descartes.dql.models.mapping.domain.AbstractSeriesResult;
import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.Result;
import tools.descartes.dql.models.mapping.domain.SeriesResultElement;
import tools.descartes.dql.models.mapping.domain.StatType;
import tools.descartes.dql.models.mapping.domain.impl.DecimalResultImpl;
import tools.descartes.dql.models.mapping.domain.impl.StatTypeImpl;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

/**
 * Interpreter for Service Level Agreements (SLA)
 * 
 * @author Jurgen Walter <juergen.walter@uni-wuerzburg.de>
 */
public class SLAInterpreter extends AbstractQueryInterpreter<Goal, PerformanceMetricsQueryConnector> {
	private static boolean isQuantitativeAnalysis = true;
	private HashMap<Sla, Double> slaPenaltiesMap;
	private HashMap<Slo, Double> sloPenaltiesMap;
	private HashMap<Slo, Boolean> sloComplianceMap;
	private HashMap<Sla, Boolean> slaComplianceMap;
	private static String constraintClause;
	private static String variationClause;

	public SLAInterpreter(final Goal query) {
		super(query);
	}

	public SLAInterpreter(final Goal query, List<QueryProgressListener> listeners) {
		super((Goal) query, listeners);
	}

	@Override
	protected Class<PerformanceMetricsQueryConnector> getQueryConnectorClass() {
		return PerformanceMetricsQueryConnector.class;
	}

	@Override
	public boolean interpretQuery() {
		Goal goal = null;
		if (getQuery() instanceof Goal) {
			goal = (Goal) getQuery();
			final UsingClause usingClause = goal.getUsingClause();
			final ModelReferenceClause modelReference = usingClause.getModelReference();
			final ModelFamily family = modelReference.getFamily();
			final String familyIdentifier = family.getFamily();
			constraintClause = "";
			if (goal.getConstraintClause() != null) {
				constraintClause = "\nCONSTRAINED AS '" + goal.getConstraintClause().getType().toString() + "'\n";
			}
			extractVariationString(goal);

			setModelFamily(familyIdentifier);

			if (!initQueryConnector()) {
				String exception = "Could not load Connector for family " + getModelFamily() + ", aborting.";
				log.error(exception);
				throw new IllegalStateException(exception);
			}

			// Reset before work, allowing independent caching mechanisms in
			// the DQL Connector
			getQueryConnector().reset();

			if (goal.getQueryType().equals("ENSURE")) {
				ensureSLA(goal);
				return true;
			} else if (goal.getQueryType().equals("EVALUTATE QUALITATIVE")) {
				isQuantitativeAnalysis = false;
				return interpretGoal(goal);
			} else if (goal.getQueryType().equals("EVALUTATE QUANTITATIVELY")) {
				isQuantitativeAnalysis = true;
				return interpretGoal(goal);
			} else {
				isQuantitativeAnalysis = true;
				return interpretGoal(goal);
			}

		}
		log.error("Query type not supported");
		return false;
	}

	private String getMetricClauseString(MetricClause clause) {
		String entity = clause.getMetricReference().getEntityReference();
		String metric = clause.getMetricReference().getMetricReference();
		String statType = clause.getMetricReference().getStatTypeReference();
		return entity + "." + metric; // +"." +statType
	}

	void ensureSLA(Goal goal) {
		sloComplianceMap = new HashMap<Slo, Boolean>();
		sloPenaltiesMap = new HashMap<Slo, Double>();
		EList<Slo> slos = goal.getSloQuery().getSlos();
		// Create performance metrics query
		String queryString = getQueryAsString();
		String usingAndForStatement = queryString.split("FOR")[1];
		queryString = "SELECT ";
		for (int i = 0; i < slos.size() - 1; i++) {
			Slo slo = slos.get(i);
			queryString += "\n" + getMetricClauseString((MetricClause) slo.getMetricClauseType()) + ",\n";
		}
		Slo slo = slos.get(slos.size() - 1);
		queryString += "\n" + getMetricClauseString((MetricClause) slo.getMetricClauseType()) + "\n";

		queryString += constraintClause + variationClause + " FOR " + usingAndForStatement;
		queryString = queryString.replace("%20", " ");
		queryString = queryString.replace(" @ ", "@");
		queryString += "\n";
		System.out.println("-------------");
		System.out.println(queryString);
		System.out.println("-------------");
		final DescartesQL queryModel = stringToModel(queryString);
		EObject genQuery = queryModel.getQuery();

		// Reuse of Performance Metrics connector
		PerformanceMetricsQueryInterpreter interpreter = new PerformanceMetricsQueryInterpreter(
				(PerformanceMetricsQuery) genQuery);
		boolean validQuery = interpreter.interpretQuery();
		if (!validQuery) {
			log.error("query is not valid");
		}
		List<EntityMapping> mappings = interpreter.getResults();
		double costBestSolution = Double.MAX_VALUE;
		String bestSolution = "";
		for (EntityMapping mapping : mappings) {
			double costs = getCosts(mapping);
			if (costs < costBestSolution) {
				log.info("Analyzing SLAs for " + mapping.getDescription());
				boolean isSatisfied = true;
				for (Slo slo2 : slos) { // check all SLOs, should be SLAs
					Result result = mapping.getEntities().get(0).getProbes().get(0).getStatTypes().get(0).getResult();
					isSatisfied = isSloSatisfied(slo2, result);
					if (!isSatisfied) {
						continue;
					}
				}
				if (isSatisfied) {
					log.info("new best solution " + mapping.getDescription());
					bestSolution = mapping.getDescription();
					costBestSolution = costs;
				}
			} else {
				log.info("already found cheaper solution than " + mapping.getDescription());
			}
		}
		log.info("---------------");
		if ("".equals(bestSolution)) {
			log.info("No configuration satisfied SLAs");
		} else {
			log.info("bestSolution " + bestSolution);
		}
		log.info("---------------");

		// for (dof:dofs){
		// analyzeQualitatively()
		// if()
		// interval casting (median)
		// }
	}

	private double getCosts(EntityMapping mapping) {
		int costFactor = 1;
		if (mapping.getDescription().contains("requests") | mapping.getDescription().contains("users")) {
			// workload, number of users
			costFactor = -1;
		} else if (mapping.getDescription().contains("processing rate")) {
			costFactor = 1;
		} else if (mapping.getDescription().contains("number of cores"))
			costFactor = 1;
		// cores ==> positive
		String costsString = getNumber(mapping.getDescription());
		double costs = costFactor * Double.parseDouble(costsString);
		return costs;
	}

	private String getNumber(String str) {
		StringBuffer sBuffer = new StringBuffer();
		Pattern p = Pattern.compile("[0-9]+.[0-9]*|[0-9]*.[0-9]+|[0-9]+");
		Matcher m = p.matcher(str);
		while (m.find()) {
			sBuffer.append(m.group());
		}
		return sBuffer.toString();
	}

	/**
	 * Entry point for interpreting SELECT
	 * 
	 * @param goalQuery
	 *            The query to be interpreted
	 * @return True, if interpreting the query was successful
	 */
	private boolean interpretGoal(final Goal goalQuery) {
		final UsingClause usingClause = goalQuery.getUsingClause();
		final ModelReferenceClause modelReference = usingClause.getModelReference();
		final ModelLocation model = modelReference.getModel();
		final String location = model.getLocation();

		sloPenaltiesMap = new HashMap<Slo, Double>();
		slaPenaltiesMap = new HashMap<Sla, Double>();
		sloComplianceMap = new HashMap<Slo, Boolean>();
		slaComplianceMap = new HashMap<Sla, Boolean>();

		EList<Sla> slas = goalQuery.getSlaQuery().getSlas();
		for (Sla sla : slas) {
			log.info("Evaluate " + sla.getName());
			isSlaSatisfied(location, sla);
		}

		SLAReporter.printResults(slaPenaltiesMap, sloPenaltiesMap, sloComplianceMap, slaComplianceMap, slas,
				goalQuery.getUsingClause().getModelReference().getModel().getLocation().toString());
		return !slaComplianceMap.containsValue(false);
	}

	private EntityMapping triggerPerformanceMetricsConnector(MetricClauseType ref, String modelLocation) {
		if (ref instanceof MetricClause) {
			String entity = ((MetricClause) ref).getMetricReference().getEntityReference();
			String metric = ((MetricClause) ref).getMetricReference().getMetricReference();
			String statType = ((MetricClause) ref).getMetricReference().getStatTypeReference();

			// Create performance metrics query
			String queryString = getQueryAsString();
			queryString = queryString.split("FOR")[1];
			System.out.println(constraintClause);
			queryString = "SELECT " + entity + "." + metric // + "." + statType
					+ constraintClause + variationClause + " FOR" + queryString;
			queryString = queryString.replace("%20", " ");
			queryString = queryString.replace(" @ ", "@");
			queryString += "\n";
			final DescartesQL queryModel = stringToModel(queryString);
			EObject genQuery = queryModel.getQuery();

			// Reuse of Performance Metrics connector
			PerformanceMetricsQueryInterpreter interpreter = new PerformanceMetricsQueryInterpreter(
					(PerformanceMetricsQuery) genQuery);
			boolean validQuery = interpreter.interpretQuery();
			if (!validQuery) {
				log.error("query is not valid");
			}
			EntityMapping entityMapping = interpreter.getResult();
			return entityMapping;
		}
		return null;
	}

	private boolean isSlaSatisfied(final String location, Sla sla) {
		log.info(sla.getName());
		for (WeightedSlo wslo : sla.getWeightedSlo()) {
			Slo slo = wslo.getSlo();
			int weight = 1;
			// try {
			// weight = wslo.getWeight();
			// } catch (Exception e) {
			// log.info("could not parse weigth ==> set weight to 1");
			// weight = 1;
			// }
			if (!sloComplianceMap.containsKey(slo)) {
				boolean isSloSatisfied = isSloSatisfied(slo, location);
				if (!isSloSatisfied) {
					slaComplianceMap.put(sla, false);
					if (isQuantitativeAnalysis) {
						double penalty = sloPenaltiesMap.get(slo);
						if (!slaPenaltiesMap.containsKey(sla)) {
							slaPenaltiesMap.put(sla, weight * penalty);
						} else {
							slaPenaltiesMap.put(sla, slaPenaltiesMap.get(sla) + weight * penalty);
						}
					} else {
						return false;
					}
				}
			} else {
				if (!sloComplianceMap.get(slo)) {
					if (!slaPenaltiesMap.containsKey(sla)) {
						slaPenaltiesMap.put(sla, weight * sloPenaltiesMap.get(slo));
					} else {
						slaPenaltiesMap.put(sla, slaPenaltiesMap.get(sla) + weight * sloPenaltiesMap.get(slo));
					}
				}
			}
		}
		if (!slaComplianceMap.containsKey(sla)) {
			slaComplianceMap.put(sla, true);
			log.info(sla.getName() + " is satisfied");
		}
		return slaComplianceMap.get(sla);
	}

	private boolean isSloSatisfied(Slo slo, String modelLocation) {
		double threshold = parseThreshold(slo);
		EntityMapping entityMapping = triggerPerformanceMetricsConnector(slo.getMetricClauseType(), modelLocation);
		for (Entity m : entityMapping.getEntities()) {
			for (Probe p : m.getProbes()) {
				DecimalResult r = new DecimalResultImpl(threshold);
				StatType s = new StatTypeImpl(r);
				s.setTypeName("Threshold SLO " + slo.getName());
				p.getStatTypes().add(s);
			}
		}
		// this.getResults().add(entityMapping);
		this.setResult(entityMapping);
		Result result = entityMapping.getEntities().get(0).getProbes().get(0).getStatTypes().get(0).getResult();

		return isSloSatisfied(slo, result);
	}

	private boolean isSloSatisfied(Slo slo, Result result) {
		double threshold = parseThreshold(slo);

		boolean isSatisfied = true;
		double penalty = 0;
		String comparator = slo.getComparator().getLiteral();
		if (result instanceof DecimalResult) {
			isSatisfied = isSatisfied((DecimalResult) result, comparator, threshold);
			log.info(slo.getName() + ": " + ((DecimalResult) result).getValue().doubleValue() + " " + comparator + " "
					+ threshold + " ==> " + isSatisfied);
			if (isQuantitativeAnalysis && !isSatisfied) {
				try {
					penalty = slo.getPen().getViolation().getAmount();
				} catch (NullPointerException e) {
					penalty = 0;
				}
			}
		} else if (result instanceof AbstractSeriesResult) {
			EList<SeriesResultElement> elements = ((AbstractSeriesResult) result).getElements();
			int violations = countViolations(elements, comparator, threshold);
			double violationThreshold = 1;
			try {
				violationThreshold = slo.getPen().getViolation().getAmount();
			} catch (NullPointerException e) {
			}

			String currency = "";
			try {
				currency = slo.getPen().getCurrency().toString();
			} catch (NullPointerException e) {
			}

			if (Math.floor(violations / violationThreshold) > 0) {
				isSatisfied = false;
			} else {
				isSatisfied = true;
			}
			if (isQuantitativeAnalysis) {
				double violationValue = 0;
				try {
					violationValue = slo.getPen().getValue();
				} catch (NullPointerException e) {
				}
				penalty = Math.floor(violations / violationThreshold) * violationValue;
				log.info(slo.getName());
				log.info("\tnumber of violations: " + violations);
				log.info("\tpercentage of violations: " + (double) violations * 100 / elements.size() + "%");
				log.info("\tnumber of violations units: " + Math.floor(violations / violationThreshold));
				log.info("\tpenalty: " + penalty + " " + currency);
			}
		} else {
			isSatisfied = false;
			log.info("Could not evaluate slo" + slo.getName());
			log.info("Only decimal and abstractSeries results supported");
		}
		sloComplianceMap.put(slo, isSatisfied);
		sloPenaltiesMap.put(slo, penalty);
		return isSatisfied;
	}

	private double parseThreshold(Slo slo) {
		double threshold = slo.getValue().getDoubleValue();
		if (slo.getValue().getIntValue() > threshold) {
			threshold = slo.getValue().getIntValue();
		}
		return threshold;
	}

	private boolean isSatisfied(DecimalResult result, String comparator, double threshold) {
		// TODO implement with comparator
		boolean isSatisfied = false;
		if (comparator.equals(">=")) {
			isSatisfied = result.getValue().doubleValue() >= threshold;
		} else if (comparator.equals(">")) {
			isSatisfied = result.getValue().doubleValue() > threshold;
		} else if (comparator.equals("<")) {
			isSatisfied = (result).getValue().doubleValue() < threshold;
		} else if (comparator.equals("<=")) {
			isSatisfied = (result).getValue().doubleValue() <= threshold;
		}
		return isSatisfied;
	}

	private int countViolations(EList<SeriesResultElement> elements, String comparator, double threshold) {
		int violations = 0;
		for (SeriesResultElement element : elements) {
			double value = element.getY().doubleValue();
			if (comparator.equals(">=")) {
				if (value < threshold)
					violations++;
			} else if (comparator.equals(">")) {
				if (value <= threshold)
					violations++;
			} else if (comparator.equals("<")) {
				if (value >= threshold)
					violations++;
			} else if (comparator.equals("<=")) {
				if (value > threshold)
					violations++;
			}
		}
		return violations;
	}

	/**
	 * Process a CONSTRAINED AS clause in best-effort manner
	 * 
	 * @param queryConnector
	 *            The query connector to execute the query
	 * @param explorationStrategyClause
	 */
	private void processExplorationStrategyClause(final PerformanceMetricsQueryConnector queryConnector,
			final ExplorationStrategyClause explorationStrategyClause) {
		final ConfigurationPropertiesClause configurationProperties = explorationStrategyClause.getParameterClause();
		final Properties explorationStrategyProperties = toFlatConfigurationPropertiesClause(configurationProperties);
		queryConnector.setDoFExplorationStrategy(explorationStrategyClause.getName(), explorationStrategyProperties);
	}

	private void extractVariationString(Goal goal) {
		variationClause = "";
		if (goal.getDofClause() != null) {
			variationClause = "\nEVALUATE DOF VARYING '" + goal.getDofClause().getVaryingClause().getMode();
			variationClause = goal.getDofClause().getMode().toString() + " " + "VARYING ";
			for (DoFReference xy : goal.getDofClause().getVaryingClause().getDofReferences()) {
				String dofStatement = "'" + xy.getDoFIdentifier().toString() + "'" + " AS "
						+ xy.getDoFAliasClause().getAlias();
				dofStatement += " <";
				DoFVariationClause dofVariationClause = ((DoFVariationClause) (xy.getDoFVariationClause()));

				EObject variationClause2 = dofVariationClause.getVariation();
				if (variationClause2 instanceof ValueVariationClause) {
					ValueVariationClause ab = (ValueVariationClause) variationClause2;
					for (int i = 0; i < ab.getValue().size() - 1; i++) {
						dofStatement += ab.getValue().get(i) + ",";
					}
					dofStatement += ab.getValue().get(ab.getValue().size() - 1);

				} else if (variationClause2 instanceof IntervalVariationClause) {
					IntervalVariationClause ab = (IntervalVariationClause) variationClause2;
					dofStatement += ((IntervalVariationClause) variationClause2).getStart() + " ... "
							+ ((IntervalVariationClause) variationClause2).getEnd();
					dofStatement += " BY " + ((IntervalVariationClause) variationClause2).getStep();
				} else if (variationClause2 instanceof DoubleValueVariationClause) {
					DoubleValueVariationClause ab = (DoubleValueVariationClause) variationClause2;
					for (int i = 0; i < ab.getValue().size() - 1; i++) {
						dofStatement += ab.getValue().get(i) + ",";
					}
					dofStatement += ab.getValue().get(ab.getValue().size() - 1);
				}
				dofStatement += ">\n";
				variationClause += dofStatement;

			}
		}
	}

}
