/**
 * ==============================================
 *  DQL : Descartes Query Language
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Juergen Walter, Fabian Gorsler, Fabian Brosig and other contributors.
 *
 * Project Info:   http://descartes.tools/dql
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.core.engine.whatif;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.eclipse.emf.ecore.util.EcoreUtil;

import tools.descartes.dql.connector.ModelStructureQueryConnector;
import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;
import tools.descartes.dql.core.engine.interpreter.QueryInterpreter;
import tools.descartes.dql.core.engine.util.DQLLogger;
import tools.descartes.dql.lang.descartesQL.WhatIfQuery;
import tools.descartes.dql.lang.descartesQL.WhatIfVariationClause;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.StatType;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

public class WhatIfQueryInterpreter implements QueryInterpreter {

	private static final Logger logger = DQLLogger.getLogger(WhatIfQueryInterpreter.class.getName());

	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;

	private List<EntityMapping> results = new LinkedList<EntityMapping>();

	private ModelStructureQueryConnector modelStructureQuerConnector = null;
	private PerformanceMetricsQueryConnector performanceMetricsConnector = null;

	// Connector State
	private WhatIfQuery query;
	private String modelLocation;
	private List<QueryProgressListener> listeners = new ArrayList<QueryProgressListener>();

	// constructor
	public WhatIfQueryInterpreter(final WhatIfQuery query) {
		this.query = query;
		this.modelLocation = query.getUsingClause().getModelReference().getModel().getLocation();
	}

	// constructor
	public WhatIfQueryInterpreter(final WhatIfQuery query, final List<QueryProgressListener> listeners) {
		this.query = query;
		this.listeners = listeners;
		this.modelLocation = query.getUsingClause().getModelReference().getModel().getLocation();
	}

	@Override
	public EntityMapping getResult() {
		if (results == null) {
			return null;
		}

		return results.get(0);
	}

	@Override
	public List<EntityMapping> getResults() {
		return results;
	}

	@Override
	public boolean hasMultipleResults() {
		return true;
	}

	@Override
	public boolean interpretQuery() {

		/*
		 * no need to split interpretation into sub-mehtods like
		 * PerformanceMetricsQueryInterpreter.
		 * 
		 * QueryExecutionEngineImpl checks if the query is a WhatIfQuery or not
		 * => no need to check again. Called methods are checking step by step
		 * if query input is valid or not.
		 * 
		 */

		initializeConnectors(query.getUsingClause().getModelReference().getFamily().getFamily());

		validateDoFs();

		EntityMapping requestWithoutDoF = createRequestWithoutDof();
		EntityMapping requestWithDoF = addDoFToRequest(requestWithoutDoF);

		// intercept results for filtering and customize them
		this.results = performanceMetricsConnector.requestDoFMetrics(requestWithDoF);

		if (query.getFilteringClause() != null)
			filterResults();
		if (query.getDeltaClause() != null)
			modifyToDeltas();

		return true;
	}

	/*
	 * modifies results that results.get(i) for i > 0 has delta-values as result
	 */
	private void modifyToDeltas() {
		EntityMapping initial = results.get(0);
		List<EntityMapping> filteredResults = new LinkedList<EntityMapping>();
		for (EntityMapping result : results)
			filteredResults.add(WhatIfUtil.createDeltas(initial, result));

	}

	/*
	 * processes the filtering
	 */
	private void filterResults() {
		WhatIfFilter filter = WhatIfFilterFactory.parseFilter(query.getFilteringClause());
		EntityMapping initial = results.get(0);
		List<EntityMapping> filteredResults = new LinkedList<EntityMapping>();
		for (EntityMapping result : results)
			filteredResults.add(filter.applyFilter(initial, result));

	}

	/*
	 * initializes a ModelStructureQueryConnector- and a
	 * PerformanceMetricsConnector instance. Both are needed for executing a
	 * WhatIfQuery.
	 */
	private void initializeConnectors(String connectorFamilyName) {
		this.modelStructureQuerConnector = WhatIfUtil.getMsqc(connectorFamilyName);
		this.performanceMetricsConnector = WhatIfUtil.getPmqc(connectorFamilyName);
		for (QueryProgressListener listener : listeners)
			this.performanceMetricsConnector.addQueryProgressListener(listener);
	}

	/*
	 * creates DoFs for executing with the PerformanceMetricsQueryConnector
	 */
	private EntityMapping addDoFToRequest(final EntityMapping requestWithoutDoF) {
		EntityMapping requestWithDoF = EcoreUtil.copy(requestWithoutDoF);
		HashMap<String, List<String>> dofIntervals = WhatIfUtil.processWhatIfVariation(query, getInitialValues());

		for (Entry<String, List<String>> entry : dofIntervals.entrySet()) {
			DoF dof = domainFactory.createDoF();
			dof.setIdentifier(entry.getKey());
			StringBuilder builder = new StringBuilder("[");

			for (String value : entry.getValue())
				builder.append(value + ", ");
			builder.delete(builder.length() - 2, builder.length());
			builder.append("]");

			dof.setProperties(new Properties());
			dof.getProperties().put("value", builder.toString());
			dof.getProperties().put("function", "fixed");

			requestWithDoF.getDoF().add(dof);
		}
		return requestWithDoF;
	}

	/*
	 * get initial values from ModelStructureQueryConnector. Connector gets an
	 * empty EntityMapping to add the DoFs with their initialValues. Those are
	 * put into the HashMap
	 */
	private HashMap<String, String> getInitialValues() {
		HashMap<String, String> initialValues = new HashMap<String, String>();
		EntityMapping mapping = modelStructureQuerConnector
				.requestDoF(WhatIfUtil.createEmptyEntityMapping(modelLocation));

		for (DoF doF : mapping.getDoF()) {
			if (null != doF.getCurrentState())
				initialValues.put(doF.getIdentifier(), doF.getCurrentState());
			else
				throw new IllegalStateException(
						"Connector did not return valid current value for DoF " + doF.getIdentifier());
		}
		return initialValues;
	}

	/*
	 * creates an EntityMapping with requested metrics. If there are no special
	 * metrics requested, all metrics the connector provieds are used
	 */
	private EntityMapping createRequestWithoutDof() {
		if (query.getMetricReferenceClauses() != null)
			return getEntityMappingForSpecificMetrics();
		else
			return getEntityMappingForAllMetrics();
	}

	/*
	 * creates an EntityMapping for all metrics the ModelStructureQueryConnector
	 * provifdes
	 */
	private EntityMapping getEntityMappingForAllMetrics() {
		EntityMapping mapping = WhatIfUtil.createEmptyEntityMapping(modelLocation);
		mapping = modelStructureQuerConnector.requestEntities(mapping);
		mapping = modelStructureQuerConnector.requestProbes(mapping);
		addDefaultResultType(removeDerivableMetrics(mapping));
		return mapping;
	}

	/**
	 * removes derivable for SELECT *
	 * 
	 * @param mapping
	 * @return mapping
	 */
	private EntityMapping removeDerivableMetrics(EntityMapping mapping) {
		for (Iterator<Entity> i = mapping.getEntities().iterator(); i.hasNext();) {
			Entity r = i.next();
			for (Iterator<Probe> i2 = r.getProbes().iterator(); i2.hasNext();) {
				Probe probe = i2.next();
				if (probe.getMetricName().contains("Perc") || probe.getMetricName().contains("avg")
						|| probe.getMetricName().contains("Histogram")) {
					logger.info("Metric " + probe.getMetricName() + " was removed for SELECT *");
					i2.remove();
				}
			}
			if (r.getProbes().isEmpty())
				i.remove();
		}
		return mapping;
	}

	private void addDefaultResultType(EntityMapping entityMapping) {
		for (Entity e : entityMapping.getEntities())
			for (Probe p : e.getProbes())
				if (p.getStatTypes().isEmpty()) {
					StatType defa = DomainFactory.eINSTANCE.createStatType();
					defa.setTypeName("DEFAULT");
					p.getStatTypes().add(defa);
				}
	}

	/*
	 * created an entityMapping for special metrics user chose in the query.
	 */
	private EntityMapping getEntityMappingForSpecificMetrics() {
		Map<String, Entity> aliasMap = new HashMap<String, Entity>();
		EntityMapping mapping = WhatIfUtil.createEmptyEntityMapping(modelLocation);
		WhatIfUtil.processForClause(query, aliasMap, mapping);
		WhatIfUtil.processSelectClause(query, aliasMap);
		return mapping;
	}

	/*
	 * Method to check if a DoF is supported or not
	 */
	private void validateDoFs() {
		List<String> queryDoFNames = requestQueryDoFNames();
		List<String> existingDoFNames = requestExistingDoFNames();
		for (String identifier : queryDoFNames)
			if (!existingDoFNames.contains(identifier))
				throw new IllegalArgumentException("The DoF " + identifier + " does not exist in SuD, aborting!");
	}

	/*
	 * getting all DoFs from query in order to compare query-DoFs with existing
	 * DoFs
	 */
	private List<String> requestQueryDoFNames() {
		List<String> dofNames = new LinkedList<String>();
		for (WhatIfVariationClause clause : query.getWhatIfVariation().getWhatIfVariationClause())
			dofNames.add(clause.getVariationPoint());
		return dofNames;
	}

	/*
	 * getting all DoFs from connector in order to compare query-DoFs with
	 * existing DoFs
	 */
	private List<String> requestExistingDoFNames() {
		List<String> existingDoFNames = new LinkedList<String>();
		EntityMapping mapping = modelStructureQuerConnector
				.requestDoF(WhatIfUtil.createEmptyEntityMapping(modelLocation));
		for (DoF dof : mapping.getDoF())
			existingDoFNames.add(dof.getIdentifier());
		return existingDoFNames;
	}

}
