/**
 * ==============================================
 *  DQL : Descartes Query Language
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Juergen Walter, Fabian Gorsler, Fabian Brosig and other contributors.
 *
 * Project Info:   http://descartes.tools/dql
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.core.registry;

import java.util.Set;

import org.osgi.framework.ServiceReference;

import tools.descartes.dql.connector.ConnectorProvider;

/**
 * Interface of of a central registry of query connectors for specific
 * performance meta-model families.
 * 
 * The implementation is based on OSGi Declarative Services as described by the
 * OSGi Standards. The usage scenario for this service is to provide a central
 * entry point in the OSGi run-time to search for query connectors for a
 * particular performance meta-model family. The query connectors are referenced
 * through OSGi Service References. Without this central data structure, a class
 * requiring a query connector would have to instantiate each query connector to
 * check for its support of a specific family.
 * 
 * @see META-INF/component.xml
 * @author Fabian Gorsler <fabian.gorsler@student.kit.edu>
 */
public interface ConnectorRegistry {
	/**
	 * Method used by all classes that implement
	 * {@link tools.descartes.dql.connector.ConnectorProvider} to register
	 * with the connector registry
	 * 
	 * @see tools.descartes.dql.connector.ConnectorProvider
	 * @param serviceReference
	 */
	public void bindConnector(ServiceReference<ConnectorProvider> serviceReference);

	/**
	 * Returns a set of names for registered connectors
	 * 
	 * @return The set of registered connector names
	 */
	public Set<String> getRegisteredConnectorNames();

	/**
	 * Method used by the query execution engine to create a reference to a
	 * query connector for a specific performance meta-model family
	 * 
	 * @see edu.kit.ipd.descartes.ql.core.engine.query.QueryExecutionEngine
	 * @param familyIdentifier
	 *            The identifier of the performance meta-model family
	 * @param queryConnector
	 *            The query class for which a query connector is required
	 * @return
	 */
	public <S, T> ServiceReference<S> requestServiceReference(String familyIdentifier, Class<T> queryConnector);

	/**
	 * Method used by all classes that implement
	 * {@link tools.descartes.dql.connector.ConnectorProvider} to
	 * unregister with the connector registry
	 * 
	 * @see tools.descartes.dql.connector.ConnectorProvider
	 * @param serviceReference
	 */
	public void unbindConnector(ServiceReference<ConnectorProvider> serviceReference);

	/**
	 * Method used by all classes that implement
	 * {@link tools.descartes.dql.connector.ConnectorProvider} to update
	 * the connector registry
	 * 
	 * @see tools.descartes.dql.connector.ConnectorProvider
	 * @param serviceReference
	 */
	public void updateConnector(ServiceReference<ConnectorProvider> serviceReference);
}
